#pragma once

#include <vector>

#include <dune/geometry/quadraturerules.hh>
#include <dune/geometry/type.hh>

#include <amdis/LocalBasisCache.hpp>
#include <amdis/common/TupleUtility.hpp>
#include <amdis/utility/ConcurrentCache.hpp>

namespace AMDiS
{
  namespace Impl
  {
    /// Pair of GeometryType and quadrature order and size
    template <class LocalBasisType, class Tag>
    struct EvaluatorCacheKey
    {
      using type = std::tuple<HashableGeometryType,typename LocalBasisType::Traits::DomainType>;
      type value;

      friend std::size_t hash_value(EvaluatorCacheKey const& t)
      {
        std::size_t seed = hash_value(std::get<0>(t.value));
        Dune::hash_range(seed, std::get<1>(t.value).begin(), std::get<1>(t.value).end());
        return seed;
      }

      friend bool operator==(EvaluatorCacheKey const& lhs, EvaluatorCacheKey const& rhs)
      {
        return lhs.value == rhs.value;
      }
    };
  }
} // end namespace AMDiS

DUNE_DEFINE_HASH(DUNE_HASH_TEMPLATE_ARGS(class LocalBasisType, class Tag),DUNE_HASH_TYPE(AMDiS::Impl::EvaluatorCacheKey<LocalBasisType,Tag>))

namespace AMDiS
{
  template <class LocalBasisType>
  class LocalBasisEvaluatorCache
  {
  public:
    using Traits = typename LocalBasisType::Traits;

    using DomainType = typename Traits::DomainType;
    using RangeType = typename Traits::RangeType;
    using RangeFieldType = typename Traits::RangeFieldType;
    using JacobianType = typename Traits::JacobianType;

    using QuadratureRule = Dune::QuadratureRule<RangeFieldType,Traits::dimDomain>;
    using QuadratureRules = Dune::QuadratureRules<RangeFieldType,Traits::dimDomain>;

  private:
    using ValuesKey = Impl::EvaluatorCacheKey<LocalBasisType, struct ValuesTag>;
    using GradientsKey = Impl::EvaluatorCacheKey<LocalBasisType, struct GradientsTag>;
    using Policy = tag::thread_local_policy;

  public:
    using ShapeValues = std::vector<RangeType>;
    using ShapeGradients = std::vector<JacobianType>;

    LocalBasisEvaluatorCache(LocalBasisType const& localBasis)
      : localBasis_(&localBasis)
    {}

    template <class LocalContext>
    ShapeValues const& evaluateFunction(LocalContext const& context, DomainType const& local)
    {
      ValuesKey key{typename ValuesKey::type{context.type(),local}};
      return ShapeValuesCache::get(key, [&](ShapeValues* data, ValuesKey const&)
      {
        localBasis_->evaluateFunction(local, *data);
      });
    }

    template <class LocalContext>
    ShapeGradients const& evaluateJacobian(LocalContext const& context, DomainType const& local)
    {
      GradientsKey key{typename GradientsKey::type{context.type(),local}};
      return ShapeGradientsCache::get(key, [&](ShapeGradients* data, GradientsKey const&)
      {
        localBasis_->evaluateJacobian(local, *data);
      });
    }

  private:
    using ShapeValuesCache = ConcurrentCache<ValuesKey, ShapeValues, Policy>;
    using ShapeGradientsCache = ConcurrentCache<GradientsKey, ShapeGradients, Policy>;

    LocalBasisType const* localBasis_;
  };

} // end namespace AMDiS
